//#################################################################################//
//                                                                                 //
//  Projekt: SIP-Türklingelmodul zur Anbindung an eine Fritzbox und MQTT-Broker    //
//  MCU:  ESP8266  1MB Flash                                                       //
//                                                                                 //
//#################################################################################//

//URSPRUNGSCOPYRIGHT:
/* ====================================================================
 *
 * Copyright (c) 2018 Juerge Liegner  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. Neither the name of the author(s) nor the names of any contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR(S) AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR(S) OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 // ERWEITERUNGEN:
 ====================================================================
 * edited by: Alois
 * created: 07.01.2020 16:56:53 
 *
 * changes: DTMF via SIP INFO added
 *          Two bell button inputs added for signalling
 *          Control of a door opener relay added
 *
 ====================================================================
 */
 
#define FIRMWARE  "1.54.00"


// i/o parameter
#define KLINGELTASTER_PIN                13
#define LED_PIN                           2
#define TUEROEFFNERRELAIS_PIN             5
#define JUMPER_PIN                        3
#define KLINGELUNTERBRECHERRELAIS_PIN    14

#include <ESP8266WiFi.h>
#include "PubSubClient.h"
#include "NTPClient.h"  
#include <WiFiUdp.h>
#include <ESP8266WebServer.h>
#include "Sip.h"
#include "interval.h"
#include <EEPROM.h>
#include <ArduinoJson.h>


char caSipIn[2048];
char caSipOut[2048];
uint32_t dooropenerDelay = 0;
uint32_t lastReconnectAttempt = 0;
bool startuploop = false;
bool telefone_stumm = false;
bool klingel_stumm = false;
char klingelzeiten[10][42] = { "Kein Eintrag\0", "Kein Eintrag\0", "Kein Eintrag\0", "Kein Eintrag\0", "Kein Eintrag\0", "Kein Eintrag\0", "Kein Eintrag\0", "Kein Eintrag\0", "Kein Eintrag\0", "Kein Eintrag\0" }; 
char systemlaufzeit[30] = "0 Tage, 0 Std., 0 Min.\0";
uint8_t wifi_config_mode = 0;
char IpString[20];
char topic_state[56];
char topic_klingelzeitenJSON[56];
char topic_klingelzeiten[10][56];
char topic_tueroeffner[56];
char topic_telefone_stumm[56];
char topic_klingel_stumm[56];
char topic_infoJSON[56];
char syslog[15][64] = { "\0", "\0", "\0", "\0", "\0", "\0", "\0", "\0", "\0", "\0", "\0", "\0", "\0", "\0", "\0" };
uint8_t mqtt_connects = 0;
bool mqtt_auto_disabled = false;
uint8_t bInDial = 0;

struct config_data {
  uint8_t ip[4];
  uint8_t mask[4];
  uint8_t gw[4];
  uint8_t dns[4];
  char wifissid[25];
  char wifipassword[40];
  char wifi_hostname[25];
  char ntpServerName[25];
  uint8_t dst_aktiv;
  uint8_t mqttIP[4];
  uint16_t mqttport;
  char mqtt_id[32];
  char mqtt_user[10];
  char mqtt_passwd[10];
  uint8_t mqtt_aktiv;
  char sipserver[20];
  char sipuser[20];
  char sippasswd[20];
  uint16_t sipport;
  char sipdialnr[6];
  char sipdialtext[16];
  uint8_t sipdigit;
  uint8_t einschaltdauer_tueroeffner;
  uint8_t telefon_klingeldauer;
};

struct config_data cfg;
Sip aSip(caSipOut, sizeof(caSipOut));
WiFiUDP ntpUDP;
WiFiClient espClient;
PubSubClient mqtt_client(espClient);
ESP8266WebServer server(80);
NTPClient timeClient(ntpUDP);
ADC_MODE(ADC_VCC);


void systemlog(String _systemlog) {
  strcpy(syslog[14], syslog[13]);
  strcpy(syslog[13], syslog[12]);
  strcpy(syslog[12], syslog[11]);
  strcpy(syslog[11], syslog[10]);
  strcpy(syslog[10], syslog[9]);
  strcpy(syslog[9], syslog[8]);
  strcpy(syslog[8], syslog[7]);
  strcpy(syslog[7], syslog[6]);
  strcpy(syslog[6], syslog[5]);
  strcpy(syslog[5], syslog[4]);
  strcpy(syslog[4], syslog[3]);
  strcpy(syslog[3], syslog[2]);
  strcpy(syslog[2], syslog[1]);
  strcpy(syslog[1], syslog[0]);
  strcpy(syslog[0], timeClient.getShortDateAndTime().c_str());
  strcat(syslog[0], _systemlog.c_str());
  return;
}

void set_mqttTopics() {
  strcpy(topic_state, cfg.mqtt_id);
  strcat(topic_state, "/Klingelstatus");
  strcpy(topic_klingelzeitenJSON, cfg.mqtt_id);
  strcat(topic_klingelzeitenJSON, "/Klingelzeiten/JSON");
  strcpy(topic_klingelzeiten[0], cfg.mqtt_id);
  strcat(topic_klingelzeiten[0], "/Klingelzeiten/1");
  strcpy(topic_klingelzeiten[1], cfg.mqtt_id);
  strcat(topic_klingelzeiten[1], "/Klingelzeiten/2");
  strcpy(topic_klingelzeiten[2], cfg.mqtt_id);
  strcat(topic_klingelzeiten[2], "/Klingelzeiten/3");
  strcpy(topic_klingelzeiten[3], cfg.mqtt_id);
  strcat(topic_klingelzeiten[3], "/Klingelzeiten/4");
  strcpy(topic_klingelzeiten[4], cfg.mqtt_id);
  strcat(topic_klingelzeiten[4], "/Klingelzeiten/5");
  strcpy(topic_klingelzeiten[5], cfg.mqtt_id);
  strcat(topic_klingelzeiten[5], "/Klingelzeiten/6");
  strcpy(topic_klingelzeiten[6], cfg.mqtt_id);
  strcat(topic_klingelzeiten[6], "/Klingelzeiten/7");
  strcpy(topic_klingelzeiten[7], cfg.mqtt_id);
  strcat(topic_klingelzeiten[7], "/Klingelzeiten/8");
  strcpy(topic_klingelzeiten[8], cfg.mqtt_id);
  strcat(topic_klingelzeiten[8], "/Klingelzeiten/9");
  strcpy(topic_klingelzeiten[9], cfg.mqtt_id);
  strcat(topic_klingelzeiten[9], "/Klingelzeiten/10");
  strcpy(topic_tueroeffner, cfg.mqtt_id);
  strcat(topic_tueroeffner, "/Tueroeffner");
  strcpy(topic_telefone_stumm, cfg.mqtt_id);
  strcat(topic_telefone_stumm, "/Telefonklingeln_Aus");
  strcpy(topic_klingel_stumm, cfg.mqtt_id);
  strcat(topic_klingel_stumm, "/Tuerklingel_Aus");
  strcpy(topic_infoJSON, cfg.mqtt_id);
  strcat(topic_infoJSON, "/info");
  return;
}

// MQTT Nachrichten empfangen
void callback(char* topic, byte* payload, unsigned int length) {

  payload[length] = '\0';
  String strTopic = String((char*)topic);
  String strPayload = String((char*)payload);
  if (strTopic == topic_tueroeffner) {
    if (strPayload == "true" && dooropenerDelay == 0) {
      dooropenerDelay = millis();
      digitalWrite(TUEROEFFNERRELAIS_PIN, HIGH);  //Türöffner schalten
      systemlog(" Tueroeffner betaetigt per MQTT");
    }
  }
  if (strTopic == topic_klingel_stumm) {
    if (strPayload == "true") {
      klingel_stumm = true;
    digitalWrite(KLINGELUNTERBRECHERRELAIS_PIN, HIGH);
    digitalWrite(LED_PIN, LOW);  // Setze Ausgang ledPin auf LOW / LED Ein
    systemlog(" Klingel deaktiviert per MQTT");
  } else if (strPayload == "false") {
      klingel_stumm = false;
    digitalWrite(KLINGELUNTERBRECHERRELAIS_PIN, LOW);
    digitalWrite(LED_PIN, HIGH);  // Setze Ausgang ledPin auf LOW / LED Ein
    systemlog(" Klingel aktiviert per MQTT");
  }
  }
   if (strTopic == topic_telefone_stumm) {
    if (strPayload == "true") {
    telefone_stumm = true;
    systemlog(" Telefonklingeln deaktiviert per MQTT");
  } else if (strPayload == "false") {
      telefone_stumm = false;
    systemlog(" Telefonklingeln aktiviert per MQTT");
  }
  }
  return;
}

// Mit Mqtt-Broker verbinden
boolean mqttReconnect() {
   mqtt_connects++;
  if (mqtt_connects > 200) {
    cfg.mqtt_aktiv = 0;
    mqtt_auto_disabled = true;
    systemlog(" MQTT-Broker nicht erreichbar");
    systemlog(" MQTT automatisch deaktiviert");
  }
  if (mqtt_client.connect(cfg.mqtt_id, cfg.mqtt_user, cfg.mqtt_passwd)) {
    // Once connected, publish an announcement...
    mqtt_client.subscribe(topic_tueroeffner);
    mqtt_client.subscribe(topic_klingel_stumm, 1);
    mqtt_client.subscribe(topic_telefone_stumm, 1);
    digitalWrite(LED_PIN, HIGH);
    systemlog(" MQTT-Connect");
  }
  return mqtt_client.connected();
}

void publish_klingelzeiten() {
  mqtt_client.publish(topic_klingelzeiten[0], klingelzeiten[0]);
  mqtt_client.publish(topic_klingelzeiten[1], klingelzeiten[1]);
  mqtt_client.publish(topic_klingelzeiten[2], klingelzeiten[2]);
  mqtt_client.publish(topic_klingelzeiten[3], klingelzeiten[3]);
  mqtt_client.publish(topic_klingelzeiten[4], klingelzeiten[4]);
  mqtt_client.publish(topic_klingelzeiten[5], klingelzeiten[5]);
  mqtt_client.publish(topic_klingelzeiten[6], klingelzeiten[6]);
  mqtt_client.publish(topic_klingelzeiten[7], klingelzeiten[7]);
  mqtt_client.publish(topic_klingelzeiten[8], klingelzeiten[8]);
  mqtt_client.publish(topic_klingelzeiten[9], klingelzeiten[9]);
  StaticJsonDocument<512> klingelzeiten;
  char jsonBuffer_klingelzeiten[512];
  klingelzeiten["1"] = klingelzeiten[0];
  klingelzeiten["2"] = klingelzeiten[1];
  klingelzeiten["3"] = klingelzeiten[2];
  klingelzeiten["4"] = klingelzeiten[3];
  klingelzeiten["5"] = klingelzeiten[4];
  klingelzeiten["6"] = klingelzeiten[5];
  klingelzeiten["7"] = klingelzeiten[6];
  klingelzeiten["8"] = klingelzeiten[7];
  klingelzeiten["9"] = klingelzeiten[8];
  klingelzeiten["10"] = klingelzeiten[9];
  serializeJson(klingelzeiten, jsonBuffer_klingelzeiten);
  mqtt_client.publish(topic_klingelzeitenJSON, jsonBuffer_klingelzeiten);
  return;
}


void handle_doorbellevent() {
  if (telefone_stumm == false) {
    aSip.Dial(cfg.sipdialnr, cfg.sipdialtext);  // SIP-Anwahl starten
  }
  strcpy(klingelzeiten[9], klingelzeiten[8]);
  strcpy(klingelzeiten[8], klingelzeiten[7]);
  strcpy(klingelzeiten[7], klingelzeiten[6]);
  strcpy(klingelzeiten[6], klingelzeiten[5]);
  strcpy(klingelzeiten[5], klingelzeiten[4]);
  strcpy(klingelzeiten[4], klingelzeiten[3]);
  strcpy(klingelzeiten[3], klingelzeiten[2]);
  strcpy(klingelzeiten[2], klingelzeiten[1]);
  strcpy(klingelzeiten[1], klingelzeiten[0]);
  strcpy(klingelzeiten[0], timeClient.getFullDateAndTime().c_str());
  if (klingel_stumm == true) {
    strcat(klingelzeiten[0], " [Stumm]");
  }
  if (cfg.mqtt_aktiv == 1) {
    mqtt_client.publish(topic_state, "true");
    publish_klingelzeiten();
  }
  return;
}


String getWifiQuality() {
  int16_t dbm = WiFi.RSSI();
  uint8_t quality_buffer;
  if (dbm <= -100) {
    quality_buffer = 0;
  } else if (dbm >= -50) {
    quality_buffer = 100;
  } else {
    quality_buffer = 2 * (dbm + 100);
  }
  return String(quality_buffer) + " %";
}


void publish_systeminfo() {
  StaticJsonDocument<512> info;
  char jsonBuffer_info[512];
  info["Hostname"] = cfg.wifi_hostname;
  info["Firmware"] = FIRMWARE;
  info["RSSI"] = WiFi.RSSI();
  info["Signal"] = getWifiQuality();
  info["Kanal"] = WiFi.channel();
  info["Resetgrund"] = ESP.getResetReason();
  info["Laufzeit"] = systemlaufzeit;
  info["VCC"] = (ESP.getVcc() / 1000.0);
  serializeJson(info, jsonBuffer_info);
  mqtt_client.publish(topic_infoJSON, jsonBuffer_info);
  return;
}


void update_systemlaufzeit() {
  static uint8_t minuten = 0;
  static uint8_t stunden = 0;
  static uint16_t tage = 0;
  minuten += 10;
  if (minuten == 60) {
    stunden++;
    minuten = 0;
  }
  if (stunden == 24) {
    tage++;
    stunden = 0;
  }
  strcpy(systemlaufzeit, String(tage).c_str());
  strcat(systemlaufzeit, " Tage, ");
  strcat(systemlaufzeit, String(stunden).c_str());
  strcat(systemlaufzeit, " Std., ");
  strcat(systemlaufzeit, String(minuten).c_str());
  strcat(systemlaufzeit, " Min.");
  return;
}


void setup_wifi() {
  if (wifi_config_mode == 1) {
    WiFi.softAP("config");
  } else {
    IPAddress Ip(cfg.ip[0], cfg.ip[1], cfg.ip[2], cfg.ip[3]);
    IPAddress Gw(cfg.gw[0], cfg.gw[1], cfg.gw[2], cfg.gw[3]);
    IPAddress Mask(cfg.mask[0], cfg.mask[1], cfg.mask[2], cfg.mask[3]);
    IPAddress Dns(cfg.dns[0], cfg.dns[1], cfg.dns[2], cfg.dns[3]);
    WiFi.setAutoReconnect(true);
    WiFi.persistent(false);
    WiFi.mode(WIFI_STA);
    WiFi.hostname(cfg.wifi_hostname);
    WiFi.config(Ip, Gw, Mask, Dns);
    WiFi.begin(cfg.wifissid, cfg.wifipassword);
  }
  return;
}


void setup() {
  pinMode(KLINGELTASTER_PIN, INPUT_PULLUP);
  pinMode(JUMPER_PIN, INPUT_PULLUP);
  pinMode(LED_PIN, OUTPUT);
  digitalWrite(LED_PIN, HIGH);
  pinMode(TUEROEFFNERRELAIS_PIN, OUTPUT);
  digitalWrite(TUEROEFFNERRELAIS_PIN, LOW);
  pinMode(KLINGELUNTERBRECHERRELAIS_PIN, OUTPUT);
  digitalWrite(KLINGELUNTERBRECHERRELAIS_PIN, LOW);
  if (digitalRead(JUMPER_PIN) == LOW) {  // Wenn Jumper gesetzt ist, startet der ESP im AP-Modus
    wifi_config_mode = 1;
  }
  EEPROM.begin(sizeof(cfg));
  EEPROM.get(0, cfg);
  EEPROM.end();
  //WIFI konfigurieren und starten
  setup_wifi();
  //initialisiere SIP-Client
  strcpy(IpString, (String(cfg.ip[0]) + "." + String(cfg.ip[1]) + "." + String(cfg.ip[2]) + "." + String(cfg.ip[3])).c_str());
  aSip.Init(cfg.sipserver, cfg.sipport, IpString, cfg.sipport, cfg.sipuser, cfg.sippasswd, cfg.telefon_klingeldauer);
  // Internen Pullup von unbenutzten GPIOS aktivieren
  pinMode(0, INPUT_PULLUP);
  pinMode(1, INPUT_PULLUP);
  pinMode(3, INPUT_PULLUP);
  pinMode(4, INPUT_PULLUP);
  pinMode(16, INPUT_PULLUP);
  timeClient.setServerName(cfg.ntpServerName);
  timeClient.begin();
  timeClient.setUpdateInterval(3600000);  //ms
  if (cfg.dst_aktiv == 1) {
    timeClient.setTimeOffset(7200);
  } else {
    timeClient.setTimeOffset(3600);
  }
  systemlog(" Systemstart");
  systemlog(" " + ESP.getResetReason());
  //MQTT-Client initialisieren
  if (cfg.mqtt_aktiv == 1) {
    set_mqttTopics();
    IPAddress mqttsrv(cfg.mqttIP[0], cfg.mqttIP[1], cfg.mqttIP[2], cfg.mqttIP[3]);
    mqtt_client.setServer(mqttsrv, cfg.mqttport);
    mqtt_client.setCallback(callback);
    mqtt_client.setBufferSize(512);
  }
  setup_webserver();
}

//---------------------------------------------------------------------------------------------------------------------------

void loop() {

  timeClient.update();
  server.handleClient();
  
  //-------------------------------------------------------------
  int packetSize = aSip.Udp.parsePacket();
  if (packetSize > 0) {
    caSipIn[0] = 0;
    packetSize = aSip.Udp.read(caSipIn, sizeof(caSipIn));
    if (packetSize > 0) {
      caSipIn[packetSize] = 0;
    }
  }
  aSip.HandleUdpPacket((packetSize > 0) ? caSipIn : 0);

  //-------------------------------------------------------------
  // Bei Verbindungsverlust mit MQTT-Broker neu verbinden
  if (cfg.mqtt_aktiv == 1) {
    if (!mqtt_client.connected()) {
      uint32_t now = millis();
      if (now - lastReconnectAttempt > 5000) {
        lastReconnectAttempt = now;
        // Attempt to reconnect
        if (mqttReconnect()) {
          lastReconnectAttempt = 0;
          mqtt_connects = 0;
        }
      }
    } else {
      // Client connected
      mqtt_client.loop();
    }
  }
  //-------------------------------------------------------------

  if (startuploop == false && cfg.mqtt_aktiv == 1 && millis() > 15000) {
    mqtt_client.publish(topic_state, "false");
    publish_systeminfo();
    publish_klingelzeiten();
    startuploop = true;
  }
  //------------------------------------------------------------------
  if (!mqtt_client.connected() && wifi_config_mode == 0 && cfg.mqtt_aktiv == 1) {
    INTERVAL(1000UL) {
      digitalWrite(LED_PIN, !digitalRead(LED_PIN));
    }
  }

  if (!mqtt_client.connected() && wifi_config_mode == 1) {
    INTERVAL(300UL) {
      digitalWrite(LED_PIN, !digitalRead(LED_PIN));
    }
  }

  //=================================================================================================================
  // Abfrage der Klingeltaster
  static uint8_t buttonState;
  buttonState = digitalRead(KLINGELTASTER_PIN);  // Status von Eingang KLINGELTASTER_PIN wird an die Variable buttonState übergeben
  

  if (buttonState == LOW)  // Wenn Klingeltaster gedrückt...
  {
   // digitalWrite(LED_PIN, LOW);  // Setze Ausgang ledPin auf LOW / LED Ein
    if (!bInDial) {
      handle_doorbellevent();  //
      bInDial = 1;
    }
  }

  if (buttonState == HIGH)  // Wenn Klingeltaster wieder losgelassen....
  {
    if (!aSip.IsBusy()) {
      bInDial = 0;
      //digitalWrite(LED_PIN, HIGH);  //Setze Ausgang ledPin auf LOW / LED Aus
    }
  }
  // Abfrage der Klingeltaster Ende
  //=================================================================================================================



  //=================================================================================================================
  // Ansteuerung Türöffner

  //Türöffnerralais schalten per Telefon
  if (aSip.iSignal == cfg.sipdigit && dooropenerDelay == 0) {
    dooropenerDelay = millis();
    digitalWrite(TUEROEFFNERRELAIS_PIN, HIGH);
    systemlog(" Tueroeffner betaetigt per Telefon");
  }

  //Türöffnerralais nach ....ms (dooropenertime) wieder abschalten
  if (dooropenerDelay && (millis() - dooropenerDelay) > (cfg.einschaltdauer_tueroeffner * 1000)) {
    dooropenerDelay = 0;
    aSip.iSignal = 0;
    digitalWrite(TUEROEFFNERRELAIS_PIN, LOW);  //Setze Ausgang TUEROEFFNERRELAIS_PIN auf LOW
    if (cfg.mqtt_aktiv == 1) {
      mqtt_client.publish(topic_tueroeffner, "false");  // Tueröffnertopic wieder auf false setzen
    }
  }
  // Ansteuerung Türöffner Ende
  //=================================================================================================================



  //=================================================================================================================
  // In Intervallen zu erledigende Dinge

  // Alle 10 min Laufzeit aktualisieren und Systeminfos per mqtt übertragen
  INTERVAL(600000UL) {
    update_systemlaufzeit();
    if (cfg.mqtt_aktiv == 1) {
      publish_systeminfo();
    }
  }

  // Alle 60s WiFi-Verbindung prüfen und bei Bedarf neu verbinden
  INTERVAL(60000UL) {
    if (WiFi.status() != WL_CONNECTED) {
      systemlog(" No WiFi-Connection...try to reconnect");
      WiFi.disconnect();
      WiFi.begin(cfg.wifissid, cfg.wifipassword);
    }
  }
  // In Intervallen zu erledigende Dinge ENDE
  //=================================================================================================================
}
